/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: audio_time_axis.cc,v 1.34 2003/10/26 22:05:22 essej Exp $
*/

#include <stdlib.h>
#include <math.h>

#include <algorithm>
#include <string>
#include <vector>

#include <pbd/error.h>
#include <pbd/stl_delete.h>

#include <gtkmmext/utils.h>
#include <gtkmmext/selector.h>
#include <gtkmmext/gtk_ui.h>
#include <gtkmmext/stop_signal.h>

#include <ardour/session.h>
#include <ardour/session_playlist.h>
#include <ardour/diskstream.h>
#include <ardour/utils.h>
#include <ardour/playlist.h>
#include <ardour/ladspa_plugin.h>
#include <ardour/insert.h>
#include <ardour/location.h>

#include "ardour_ui.h"
#include "editor.h"
#include "audio_time_axis.h"
#include "streamview.h"
#include "canvas-simplerect.h"
#include "plugin_manager.h"
#include "plugin_ui.h"
#include "automation_gain_line.h"
#include "automation_pan_line.h"
#include "redirect_automation_line.h"
#include "selection.h"
#include "enums.h"
#include "utils.h"

#include <ardour/audio_track.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;
using namespace LADSPA;
using namespace Gtk;

const double trim_handle_size = 6.0; /* pixels */

static const gchar * small_x_xpm[] = {
"11 11 2 1",
" 	c None",
".	c #FFFFFF",
"           ",
"           ",
"  .     .  ",
"   .   .   ",
"    . .    ",
"     .     ",
"    . .    ",
"   .   .   ",
"  .     .  ",
"           ",
"           "};

AudioTimeAxisView::AudioTimeAxisView (Editor& ed, Session& sess, Route& rt, Widget *canvas)
	: AxisView(sess),
	  RouteUI(rt, sess, _("m"), _("s"), _("r")), // mute, solo, and record
	  TimeAxisView(sess,ed,canvas),
	  button_table (3, 3),
	  edit_group_button (_("g")), // group
	  history_button (_("e")), // edit list
	  playlist_button (_("p")), 
	  loop_button (_("l")), 
	  size_button (_("h")), // height
	  automation_button (_("a")),
	  visual_button (_("v")),
	  history_selector (_refill_history_list, this, 1, true, true),
	  redirect_display (1)
{
	string history_title;

	history_mapped = false;

	playlist_menu = 0;
	playlist_action_menu = 0;
	automation_action_menu = 0;
	gain_line = 0;
	pan_line = 0;
	view = 0;
	selection_group = 0;
	timestretch_rect = 0;
	waveform_item = 0;

	selection_group = gtk_canvas_item_new (GTK_CANVAS_GROUP(canvas_display), 
					       gtk_canvas_group_get_type (), 
					       NULL);
	gtk_canvas_item_hide (selection_group);
	
	view = new StreamView (*this);

	gain_line = new AutomationGainLine ("automation gain", _session, editor, canvas_display, 
					    _route.gain_automation_curve(),
					    Editor::_canvas_control_point_event,
					    Editor::_canvas_line_event);
	gain_line->set_line_color (RGBA_TO_UINT (99,249,122,255));

	pan_line = new AutomationPanLine ("automation pan", _session, editor, canvas_display, 
					  _route.pan_automation_curve(),
					  Editor::_canvas_control_point_event,
					  Editor::_canvas_line_event);
	pan_line->set_line_color (RGBA_TO_UINT (255,213,48,255));

	history_selector.clist().set_name ("HistorySelector");
	history_selector.choice_made.connect (slot (*this, &AudioTimeAxisView::random_history_selection));

	history_title = _("ardour: ");
	history_title += _route.name();
	history_title += _(" edit history");
	
	history_display.add (history_selector);
	history_display.set_title (history_title);
	history_display.set_default_size (450, 150);
	history_display.delete_event.connect (bind (slot (Gtkmmext::UI::just_hide_it), static_cast<Gtk::Window*>(&history_display)));
	history_display.show.connect (bind (slot (*this, &AudioTimeAxisView::history_map_follower), true));
	history_display.hide.connect (bind (slot (*this, &AudioTimeAxisView::history_map_follower), false));
	
	ignore_toggle = false;

	rec_enable_button.set_active (false);
	mute_button.set_active (false);
	solo_button.set_active (false);
	
	rec_enable_button.set_name ("TrackRecordEnableButton");
	mute_button.set_name ("TrackMuteButton");
	solo_button.set_name ("SoloButton");
	edit_group_button.set_name ("TrackGroupButton");
	history_button.set_name ("TrackHistoryButton");
	playlist_button.set_name ("TrackPlaylistButton");
	automation_button.set_name ("TrackAutomationButton");
	loop_button.set_name ("TrackLoopButton");
	size_button.set_name ("TrackSizeButton");
	visual_button.set_name ("TrackVisualButton");
	hide_button.set_name ("TrackRemoveButton");

	hide_button.add (*(manage (new Pixmap (small_x_xpm))));

	_route.mute_changed.connect (slot (*this, &RouteUI::mute_changed));
	_route.solo_changed.connect (slot (*this, &RouteUI::solo_changed));
	_route.solo_safe_changed.connect (slot (*this, &RouteUI::solo_changed));
	
	solo_button.button_press_event.connect (slot (*this, &RouteUI::solo_press));
	solo_button.button_release_event.connect (slot (*this, &RouteUI::solo_release));
	mute_button.button_press_event.connect (slot (*this, &RouteUI::mute_press));
	mute_button.button_release_event.connect (slot (*this, &RouteUI::mute_release));
	rec_enable_button.button_press_event.connect (slot (*this, &RouteUI::rec_enable_press));
	edit_group_button.button_release_event.connect (slot (*this, &AudioTimeAxisView::edit_click));
	history_button.clicked.connect (slot (*this, &AudioTimeAxisView::history_click));
	playlist_button.clicked.connect (slot (*this, &AudioTimeAxisView::playlist_click));
	automation_button.clicked.connect (slot (*this, &AudioTimeAxisView::automation_click));
	loop_button.clicked.connect (slot (*this, &AudioTimeAxisView::loop_toggle));
	size_button.button_release_event.connect (slot (*this, &AudioTimeAxisView::size_click));
	visual_button.clicked.connect (slot (*this, &AudioTimeAxisView::visual_click));
	hide_button.clicked.connect (slot (*this, &AudioTimeAxisView::hide_click));

	name_entry.activate.connect (slot (*this, &AudioTimeAxisView::name_entry_activated));
	name_entry.focus_out_event.connect (slot (*this, &AudioTimeAxisView::name_entry_focus_out_handler));
	name_entry.button_press_event.connect (slot (*this, &AudioTimeAxisView::name_entry_button_press_handler));
	name_entry.button_release_event.connect (slot (*this, &AudioTimeAxisView::name_entry_button_release_handler));
	name_entry.key_release_event.connect (slot (*this, &AudioTimeAxisView::name_entry_key_release_handler));

	button_table.set_border_width (2);
	button_table.set_row_spacings (0);
	button_table.set_col_spacings (0);
	button_table.set_homogeneous (true);
	button_table.show ();

	if (is_audio_track()) {
		button_table.attach (rec_enable_button, 0, 1, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND, 0, 0);
	}
	button_table.attach (mute_button, 1, 2, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND, 0, 0);
	button_table.attach (solo_button, 2, 3, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_FILL|GTK_EXPAND, 0, 0);

	button_table.attach (loop_button, 0, 1, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND, 0, 0);
	button_table.attach (edit_group_button, 1, 2, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND, 0, 0);
	button_table.attach (history_button, 2, 3, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND, 0, 0);

	ARDOUR_UI::instance()->tooltips().set_tip(rec_enable_button, _("Record"));
	ARDOUR_UI::instance()->tooltips().set_tip(solo_button,_("Solo"));
	ARDOUR_UI::instance()->tooltips().set_tip(mute_button,_("Mute"));
	ARDOUR_UI::instance()->tooltips().set_tip(loop_button,_("Loop"));
	ARDOUR_UI::instance()->tooltips().set_tip(edit_group_button,_("Edit Group"));
	ARDOUR_UI::instance()->tooltips().set_tip(history_button,_("History"));
	ARDOUR_UI::instance()->tooltips().set_tip(size_button,_("Display Height"));
	ARDOUR_UI::instance()->tooltips().set_tip(playlist_button,_("Playlist"));
	ARDOUR_UI::instance()->tooltips().set_tip(automation_button, _("Automation"));
	ARDOUR_UI::instance()->tooltips().set_tip(visual_button, _("Visual options"));
	ARDOUR_UI::instance()->tooltips().set_tip(hide_button, _("Hide this track"));
	
	label_view ();

	// other_button_hbox.set_border_width (2);
	// other_button_hbox.set_spacing (2);
	other_button_hbox.pack_start (hide_button, false, false);
	other_button_hbox.pack_start (visual_button, false, false);
	other_button_hbox.pack_start (size_button, false, false);
	other_button_hbox.pack_end (automation_button, false, false);

	if (is_audio_track()) {
		other_button_hbox.pack_start (playlist_button, false, false);
		other_button_hbox.show_all ();
	}

	/* Add some AudioTimeAxisView specific controls to the standard LHS controls */
	name_vbox.pack_start (other_button_hbox, false, false);

	controls_hbox.pack_start (button_table, false, false);

	/* remove focus from the buttons */
	
	automation_button.unset_flags (GTK_CAN_FOCUS);
	solo_button.unset_flags (GTK_CAN_FOCUS);
	mute_button.unset_flags (GTK_CAN_FOCUS);
	edit_group_button.unset_flags (GTK_CAN_FOCUS);
	loop_button.unset_flags (GTK_CAN_FOCUS);
	history_button.unset_flags (GTK_CAN_FOCUS);
	size_button.unset_flags (GTK_CAN_FOCUS);
	playlist_button.unset_flags (GTK_CAN_FOCUS);
	hide_button.unset_flags (GTK_CAN_FOCUS);
	visual_button.unset_flags (GTK_CAN_FOCUS);

	/* map current state of the route */

	update_diskstream_display ();
	solo_changed(0);
	mute_changed(0);
	redirects_changed (0);
	gain_line->reset ();
	pan_line->reset ();
	reset_redirect_automation_curves ();

	selected = false;
	edit_group_menu_radio_group = 0;
	loop_selector = 0;
	y_position = -1;

	set_height_from_route ();
	
	_route.mute_changed.connect (slot (*this, &RouteUI::mute_changed));
	_route.solo_changed.connect (slot (*this, &RouteUI::solo_changed));
	_route.redirects_changed.connect (slot (*this, &AudioTimeAxisView::redirects_changed));

	_route.name_changed.connect (slot (*this, &AudioTimeAxisView::route_name_changed));

#if 0
	_route.redirect_automation_changed.connect (slot (*this, &AudioTimeAxisView::redirect_automation_changed));
	_route.gain_automation_curve().Changed.connect (bind (slot (*this, &AudioTimeAxisView::automation_control_point_changed), GainAutomation));
	_route.pan_automation_curve().Changed.connect (bind (slot (*this, &AudioTimeAxisView::automation_control_point_changed), PanAutomation));
#endif 
	
	if (is_audio_track()) {

		/* track */

		audio_track()->diskstream_changed.connect (slot (*this, &AudioTimeAxisView::diskstream_changed));
		get_diskstream()->speed_changed.connect (slot (*this, &AudioTimeAxisView::speed_changed));

		controls_ebox.set_name ("AudioTrackControlsBaseUnselected");
		expansion_box.set_name ("AudioTrackControlsBaseUnselected");
		controls_base_selected_name = "AudioTrackControlsBaseSelected";
		controls_base_unselected_name = "AudioTrackControlsBaseUnselected";
	} else {

		/* bus */

		controls_ebox.set_name ("BusControlsBaseUnselected");
		expansion_box.set_name ("BusControlsBaseUnselected");
		controls_base_selected_name = "BusControlsBaseSelected";
		controls_base_unselected_name = "BusControlsBaseUnselected";
	}

	editor.frames_per_unit_changed.connect (slot (*this, &AudioTimeAxisView::reset_samples_per_unit));
	editor.map_event.connect (slot (*this, &AudioTimeAxisView::when_displayed));
}

AudioTimeAxisView::~AudioTimeAxisView ()
{
	EMIT_SIGNAL GoingAway ();

	if (playlist_menu) {
		delete playlist_menu;
		playlist_menu = 0;
	}

	if (loop_selector) {
		delete loop_selector;
		loop_selector = 0;
	}

	if (gain_line) {
		delete gain_line;
		gain_line = 0;
	}

	vector_delete (&redirect_automation_curves);

	for (list<SelectionRect*>::iterator i = free_selection_rects.begin(); i != free_selection_rects.end(); ++i) {
		gtk_object_destroy (GTK_OBJECT((*i)->rect));
		gtk_object_destroy (GTK_OBJECT((*i)->start_trim));
		gtk_object_destroy (GTK_OBJECT((*i)->end_trim));
	}

	for (list<SelectionRect*>::iterator i = used_selection_rects.begin(); i != used_selection_rects.end(); ++i) {
		gtk_object_destroy (GTK_OBJECT((*i)->rect));
		gtk_object_destroy (GTK_OBJECT((*i)->start_trim));
		gtk_object_destroy (GTK_OBJECT((*i)->end_trim));
	}

	if (view) {
		delete view;
		view = 0;
	}

	if (selection_group) {
		gtk_object_destroy (GTK_OBJECT (selection_group));
		selection_group = 0;
	}

	if (canvas_display) {
		gtk_object_destroy (GTK_OBJECT (canvas_display));
		canvas_display = 0;
	}
}

gint
AudioTimeAxisView::when_displayed (GdkEventAny* ev)
{
	//set_height ((TrackHeight) height);
	return FALSE;
}

void
AudioTimeAxisView::set_playlist (Playlist *newplaylist)

{
	Playlist *pl;

	loop_set_connection.disconnect();
	modified_connection.disconnect ();
	state_changed_connection.disconnect ();
	
	if ((pl = playlist()) != 0) {
		state_changed_connection = pl->StateChanged.connect (slot (*this, &AudioTimeAxisView::playlist_state_changed));
		state_pushed_connection = pl->StatePushed.connect (slot (*this, &AudioTimeAxisView::playlist_state_pushed));
	}
}

gint
AudioTimeAxisView::edit_click (GdkEventButton *ev)
{
	if (Keyboard::modifier_state_equals (ev->state, Keyboard::Control)) {
	        _route.set_edit_group (0, this);
		return FALSE;
	} 

	using namespace Menu_Helpers;

	MenuList& items = edit_group_menu.items ();

	items.clear ();
	items.push_back (RadioMenuElem (edit_group_menu_radio_group, _("No group"), 
				   bind (slot (*this, &AudioTimeAxisView::set_edit_group_from_menu), (RouteGroup *) 0)));
	
	if (_route.edit_group() == 0) {
		static_cast<RadioMenuItem*>(items.back())->set_active ();
	}

	_session.foreach_edit_group (this, &AudioTimeAxisView::add_edit_group_menu_item);
	edit_group_menu.popup (ev->button, ev->time);

	return FALSE;
}

void
AudioTimeAxisView::add_edit_group_menu_item (RouteGroup *eg)
{
	using namespace Menu_Helpers;

	MenuList &items = edit_group_menu.items();
	items.push_back (RadioMenuElem (edit_group_menu_radio_group,
					eg->name(), bind (slot (*this, &AudioTimeAxisView::set_edit_group_from_menu), eg)));
	if (_route.edit_group() == eg) {
		static_cast<RadioMenuItem*>(items.back())->set_active ();
	}
}

void
AudioTimeAxisView::set_edit_group_from_menu (RouteGroup *eg)

{
	_route.set_edit_group (eg, this);
}

void
AudioTimeAxisView::playlist_state_changed ()

{
	Gtkmmext::UI::instance()->call_slot (slot (AudioTimeAxisView::history_selector, &Gtkmmext::Selector::rescan));
}

void
AudioTimeAxisView::playlist_state_pushed ()

{
	Gtkmmext::UI::instance()->call_slot (slot (AudioTimeAxisView::history_selector, &Gtkmmext::Selector::rescan));
}

void
AudioTimeAxisView::playlist_changed ()

{
	label_view ();

	if (is_audio_track()) {
		set_playlist (get_diskstream()->playlist());
	}
}

void
AudioTimeAxisView::loop_set (Location *loc)
{
	/* Gack. No MVC in GTK */

	ignore_toggle = true;
	if (loc && loop_button.get_active() == false) {
		loop_button.set_active (true);
	} else if (!loc && loop_button.get_active() == true) {
		loop_button.set_active (false);
	}
	ignore_toggle = false;
}

void
AudioTimeAxisView::loop_toggle ()

{
	DiskStream *ds;

	if (ignore_toggle || (ds = get_diskstream()) == 0 ) {
		return;
	}
	
	if (!loop_button.get_active()) {
		ds->set_loop (0);
		if (loop_selector) {
			loop_selector->hide_all ();
		}
	} else {
		start_loop_selection ();
	}
}

void
AudioTimeAxisView::start_loop_selection ()

{
	if (loop_selector == 0) {
		loop_selector = new Gtkmmext::PopupSelector (_("ardour: per-track loop selector"),
							     loop_selector_refill, this,
							     false);
		loop_selector->selection_made.connect (slot (*this, &AudioTimeAxisView::loop_selection_made));
		loop_selector->set_usize (400, 200);
	}

	loop_selector->show_all ();
}

void
AudioTimeAxisView::end_loop_selection ()

{
	loop_selector->hide_all ();
}

void
AudioTimeAxisView::loop_selection_made (Gtkmmext::PopupSelector *sel, 
				Gtkmmext::SelectionResult *res)
{
	DiskStream *ds;

	if (res && ((ds = get_diskstream()) != 0)) {
		Location *location = (Location *) res->data;

		if (ds->set_loop (location)) {
			return;
		}
	} 
	
	end_loop_selection ();
}

void
AudioTimeAxisView::loop_selector_refill_internal (Locations::LocationList& locations, CList& clist)
{
	Locations::LocationList::iterator i;
	gint n;
	const gchar *rowdata[1];

	for (n = 0, i = locations.begin(); i != locations.end(); ++i, ++n) {
		rowdata[0] = (*i)->name().c_str();
		clist.insert_row (n, rowdata);
		clist.rows().back().set_data (*i);
	}
}

void
AudioTimeAxisView::loop_selector_refill (CList &clist, void *arg)

{
	AudioTimeAxisView *tv = (AudioTimeAxisView *) arg;
	tv->_session.locations()->apply (*tv, &AudioTimeAxisView::loop_selector_refill_internal, clist);
}

void
AudioTimeAxisView::history_click ()

{
	if (history_display.is_visible()) {
		history_display.hide_all ();
	} else {
		history_display.show_all ();
	}
}

void
AudioTimeAxisView::history_map_follower (bool show)

{
	if (show) {
		history_mapped = true;
		history_selector.rescan ();
	} else {
		history_mapped = false;
	}
}

/* History Handling */

void
AudioTimeAxisView::_refill_history_list (CList &list, void *arg)

{
	((AudioTimeAxisView *) arg)->refill_history_list (list);
}

void
AudioTimeAxisView::really_refill_history_list (const Playlist::StateMap& states, state_id_t current)
{
	CList& clist (history_selector.clist());
	Playlist::StateMap::const_iterator i;
	const gchar *rowdata[1];
	unsigned long row;

	clist.set_column_auto_resize (0, true);
	clist.set_column_auto_resize (1, true);

	clist.freeze ();
	clist.clear ();

	for (row = 0, i = states.begin(); i != states.end(); ++i, ++row) {
		
		rowdata[0] = (*i)->operation.c_str();
		clist.insert_row (row, rowdata);
		if (current == row) {
			clist.row (row).select ();
		}
	}

	clist.thaw ();
}

void
AudioTimeAxisView::refill_history_list (CList &list)
{
	DiskStream *stream;
	Playlist *playlist;

	if ((stream = get_diskstream()) != 0 && (playlist = stream->playlist()) != 0) {
		playlist->apply_to_history (*this, &AudioTimeAxisView::really_refill_history_list);
	}
}

void
AudioTimeAxisView::random_history_selection (Gtkmmext::Selector *selector,
				     Gtkmmext::SelectionResult *res)
{
	DiskStream *stream;
	Playlist *playlist;

	if ((stream = get_diskstream()) != 0 && (playlist = stream->playlist()) != 0) {
		playlist->set_state ((state_id_t) res->row);
	}
}

void
AudioTimeAxisView::redirects_changed (void *src)
{
	setup_redirect_automation ();
}

void
AudioTimeAxisView::remove_redirect_automation_curve (Redirect *ignored, RedirectAutomationLine *ral)
{
	vector<RedirectAutomationLine*>::iterator i;

	if ((i = find (redirect_automation_curves.begin(), redirect_automation_curves.end(), ral)) != redirect_automation_curves.end()) {
		redirect_automation_curves.erase (i);
		delete ral;
	} else {
		fatal << _("programming error: redirect automation curve missing from curve list!") << endmsg;
		/*NOTREACHED*/
	}
}

void
AudioTimeAxisView::add_redirect_automation_curve (Redirect *redirect, unsigned long what)
{
	RedirectAutomationLine *ral;

	ral = new RedirectAutomationLine (redirect->describe_parameter (what), 
					  *redirect, what, _session, editor,
					  canvas_display, redirect->automation_list(), 
					  Editor::_canvas_control_point_event,
					  Editor::_canvas_line_event);

	
	ral->set_line_color (RGBA_TO_UINT (72,87,189,255));
	ral->set_height (height);
	redirect_automation_curves.push_back (ral);
	redirect->GoingAway.connect (bind (slot (*this, &AudioTimeAxisView::remove_redirect_automation_curve), ral));
	ral->show ();
	ral->queue_reset ();
}

void
AudioTimeAxisView::add_redirect_automation_curves (Redirect *redirect)
{
	set<unsigned long> s;
	RedirectAutomationLine *ral;

	redirect->what_has_automation (s);

	for (set<unsigned long>::iterator i = s.begin(); i != s.end(); ++i) {
		if ((ral = find_redirect_automation_curve (redirect, *i)) != 0) {
			ral->queue_reset ();
		} else {
			add_redirect_automation_curve (redirect, (*i));
		}
	}
}

void
AudioTimeAxisView::setup_redirect_automation ()
{
	_route.foreach_redirect (this, &AudioTimeAxisView::add_redirect_automation_curves);
}

RedirectAutomationLine *
AudioTimeAxisView::find_redirect_automation_curve (Redirect *redirect, unsigned long what)
{
	for (vector<RedirectAutomationLine *>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {
		if (&(*i)->redirect() == redirect && (*i)->port() == what) {
			return *i;
		}
	}
	return 0;
}

void
AudioTimeAxisView::redirect_automation_changed (Redirect *redirect, unsigned long port)
{
	RedirectAutomationLine *ral;

	if ((ral = find_redirect_automation_curve (redirect, port)) == 0) {
		add_redirect_automation_curve (redirect, port);
	} else {
		ral->queue_reset ();
	}
}

void
AudioTimeAxisView::label_view ()
{
	string x = _route.name();

	if (x != name_entry.get_text()) {
		name_entry.set_text (x);
		ARDOUR_UI::instance()->tooltips().set_tip (name_entry, x);
	}
}

void
AudioTimeAxisView::route_name_changed (void *src)
{
	editor.route_name_changed (this);
	label_view ();
}

void
AudioTimeAxisView::take_name_changed (void *src)

{
	if (src != this) {
		label_view ();
	}
}

void
AudioTimeAxisView::playlist_click ()
{
	// always build a new action menu
	
	if (playlist_action_menu == 0) {
		playlist_action_menu = new Menu;
	}
	
	build_playlist_menu(playlist_action_menu);
	
	playlist_action_menu->popup (1, 0);
}

void
AudioTimeAxisView::automation_click ()
{
	if (automation_action_menu == 0) {
		/* this seems odd, but the automation action
		   menu is built as part of the display menu.
		*/
		build_display_menu ();
	}
	automation_action_menu->popup (1, 0);
}

void
AudioTimeAxisView::show_timestretch (TimeSelection& ts)
{
	double x1;
	double x2;
	double y2;
	
	hide_timestretch ();
	
	if (ts.empty()) {
		return;
	}

	/* check that the time selection was made in our route, or our edit group.
	   remember that edit_group() == 0 implies the route is *not* in a edit group.
	*/

	if (!(ts.track == this || (ts.group != 0 && ts.group == _route.edit_group()))) {
		/* this doesn't apply to us */
		return;
	}

	/* ignore it if our edit group is not active */
	
	if ((ts.track != this) && _route.edit_group() && !_route.edit_group()->is_active()) {
		return;
	}

	if (timestretch_rect == 0) {
		timestretch_rect = gtk_canvas_item_new (GTK_CANVAS_GROUP(canvas_display),
							gtk_canvas_simplerect_get_type(),
							"x1", 0.0,
							"y1", 0.0,
							"x2", 0.0,
							"y2", 0.0,
							"fill_color_rgba", RGBA_TO_UINT(229,183,183,150),
							"outline_color_rgba" , RGBA_TO_UINT(100,100,100,150),
							NULL);
	}

	gtk_canvas_item_show (timestretch_rect);
	gtk_canvas_item_raise_to_top (timestretch_rect);
	
	jack_nframes_t start, end, cnt;
	
	start = ts.front().start;
	end = ts.front().end;
	cnt = end - start + 1;
	
	x1 = start / editor.frames_per_unit;
	x2 = (start + cnt - 1) / editor.frames_per_unit;
	y2 = height - 2;
	
	gtk_object_set (GTK_OBJECT(timestretch_rect), 
			"x1", x1,
			"y1", 1.0,
			"x2", x2,
			"y2", y2,
			NULL);
}

void
AudioTimeAxisView::hide_timestretch ()
{
	if (timestretch_rect) {
		gtk_canvas_item_hide (timestretch_rect);
	}
}

void
AudioTimeAxisView::reshow_selection (TimeSelection& ts)
{
	if (GTK_OBJECT_FLAGS(GTK_OBJECT(selection_group)) & GTK_CANVAS_ITEM_VISIBLE) {
		show_selection (ts);
	}
}

void
AudioTimeAxisView::show_selection (TimeSelection& ts)
{
	double x1;
	double x2;
	double y2;
	SelectionRect *rect;
	
	/* ignore it if our edit group is not active */

	if ((ts.track != this) && _route.edit_group() && !_route.edit_group()->is_active()) {
		hide_selection ();
		return;
	}

	/* check that the time selection was made in our route, or our edit group.
	   remember that edit_group() == 0 implies the route is *not* in a edit group.
	*/

	if (!(ts.track == this || (ts.group != 0 && ts.group == _route.edit_group()))) {
		/* this doesn't apply to us */
		hide_selection ();
		return;
	}
 
	if (GTK_OBJECT_FLAGS(GTK_OBJECT(selection_group)) & GTK_CANVAS_ITEM_VISIBLE) {
		while (!used_selection_rects.empty()) {
			free_selection_rects.push_front (used_selection_rects.front());
			used_selection_rects.pop_front();
			gtk_canvas_item_hide (free_selection_rects.front()->rect);
			gtk_canvas_item_hide (free_selection_rects.front()->start_trim);
			gtk_canvas_item_hide (free_selection_rects.front()->end_trim);
		}
		gtk_canvas_item_hide (selection_group);
	}

	gtk_canvas_item_show (selection_group);
	gtk_canvas_item_raise_to_top (selection_group);
	
	for (list<AudioRange>::iterator i = ts.begin(); i != ts.end(); ++i) {
		jack_nframes_t start, end, cnt;

		start = (*i).start;
		end = (*i).end;
		cnt = end - start + 1;

		rect = get_selection_rect ((*i).id);
		
		x1 = start / editor.frames_per_unit;
		x2 = (start + cnt - 1) / editor.frames_per_unit;
		y2 = height - 2;

		gtk_object_set (GTK_OBJECT(rect->rect), 
				"x1", x1,
				"y1", 1.0,
				"x2", x2,
				"y2", y2,
				NULL);
		
		// trim boxes are at the top for selections
		
		if (x2 > x1) {
			gtk_object_set (GTK_OBJECT(rect->start_trim), 
					"x1", x1,
					"y1", 1.0,
					"x2", x1 + trim_handle_size,
					"y2", 1.0 + trim_handle_size,
					NULL);
			gtk_object_set (GTK_OBJECT(rect->end_trim), 
					"x1", x2 - trim_handle_size,
					"y1", 1.0,
					"x2", x2,
					"y2", 1.0 + trim_handle_size,
					NULL);
			gtk_canvas_item_show (rect->start_trim);
			gtk_canvas_item_show (rect->end_trim);
		} else {
			gtk_canvas_item_hide (rect->start_trim);
			gtk_canvas_item_hide (rect->end_trim);
		}

		gtk_canvas_item_show (rect->rect);
		used_selection_rects.push_back (rect);
	}

	if (gain_line) {
		gain_line->show_selection ();
	}

	if (pan_line) {
		pan_line->show_selection ();
	}

	for (vector<RedirectAutomationLine*>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {
		(*i)->show_selection ();
	}
}

void
AudioTimeAxisView::hide_selection ()
{
	if (GTK_OBJECT_FLAGS(GTK_OBJECT(selection_group)) & GTK_CANVAS_ITEM_VISIBLE) {
		while (!used_selection_rects.empty()) {
			free_selection_rects.push_front (used_selection_rects.front());
			used_selection_rects.pop_front();
			gtk_canvas_item_hide (free_selection_rects.front()->rect);
			gtk_canvas_item_hide (free_selection_rects.front()->start_trim);
			gtk_canvas_item_hide (free_selection_rects.front()->end_trim);
		}
		gtk_canvas_item_hide (selection_group);

		if (gain_line) {
			gain_line->hide_selection ();
		}
		
		if (pan_line) {
			pan_line->hide_selection ();
		}
		
		for (vector<RedirectAutomationLine*>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {
			(*i)->hide_selection ();
		}
	}
}

SelectionRect *
AudioTimeAxisView::get_selection_rect (unsigned long id)
{
	SelectionRect *rect;

	/* check to see if we already have a visible rect for this particular selection ID */

	for (list<SelectionRect*>::iterator i = used_selection_rects.begin(); i != used_selection_rects.end(); ++i) {
		if ((*i)->id == id) {
			return (*i);
		}
	}

	/* ditto for the free rect list */

	for (list<SelectionRect*>::iterator i = free_selection_rects.begin(); i != free_selection_rects.end(); ++i) {
		if ((*i)->id == id) {
			free_selection_rects.erase (i);
			return (*i);
		}
	}

	/* no existing matching rect, so go get a new one from the free list, or create one if there are none */
	
	if (free_selection_rects.empty()) {

		rect = new SelectionRect;

		rect->rect = gtk_canvas_item_new (GTK_CANVAS_GROUP(selection_group),
						  gtk_canvas_simplerect_get_type(),
						  "x1", 0.0,
						  "y1", 0.0,
						  "x2", 0.0,
						  "y2", 0.0,
						  "fill_color_rgba", RGBA_TO_UINT(232,247,213,120),
						  "outline_color_rgba" , RGBA_TO_UINT(100,100,100,150),
						  NULL);
		
		
		rect->start_trim = gtk_canvas_item_new (GTK_CANVAS_GROUP(selection_group),
							    gtk_canvas_simplerect_get_type(),
							    "x1", (gdouble) 0.0,
							    "x2", (gdouble) 0.0,
							    "fill_color_rgba" , RGBA_TO_UINT(100,100,100,180),
							    "outline_color_rgba" , RGBA_TO_UINT(100,100,100,255),
							    NULL);
		
		rect->end_trim = gtk_canvas_item_new (GTK_CANVAS_GROUP(selection_group),
							  gtk_canvas_simplerect_get_type(),
							  "x1", 0.0,
							  "x2", 0.0,
							  "fill_color_rgba" , RGBA_TO_UINT(100,100,100,180),
							  "outline_color_rgba" , RGBA_TO_UINT(100,100,100,255),
							  NULL);
		free_selection_rects.push_front (rect);

		gtk_signal_connect (GTK_OBJECT(rect->rect), "event",
				    (GtkSignalFunc) Editor::_canvas_selection_rect_event,
				    &editor);
		gtk_signal_connect (GTK_OBJECT(rect->start_trim), "event",
				    (GtkSignalFunc) Editor::_canvas_selection_start_trim_event,
				    &editor);
		gtk_signal_connect (GTK_OBJECT(rect->end_trim), "event",
				    (GtkSignalFunc) Editor::_canvas_selection_end_trim_event,
				    &editor);

		gtk_object_set_data(GTK_OBJECT(rect->rect), "rect", rect);
		gtk_object_set_data(GTK_OBJECT(rect->start_trim), "rect", rect);
		gtk_object_set_data(GTK_OBJECT(rect->end_trim), "rect", rect);
	} 

	rect = free_selection_rects.front();
	rect->id = id;
	free_selection_rects.pop_front();
	return rect;
}

void
AudioTimeAxisView::set_height_from_route ()
{
	XMLProperty *prop;
	
	ensure_xml_node ();

	if ((prop = xml_node->property ("track_height")) != 0) {
		if (prop->value() == "largest") {
			set_height (Largest);
		} else if (prop->value() == "large") {
			set_height (Large);
		} else if (prop->value() == "larger") {
			set_height (Larger);
		} else if (prop->value() == "normal") {
			set_height (Normal);
		} else if (prop->value() == "smaller") {
			set_height (Smaller);
		} else if (prop->value() == "small") {
			set_height (Small);
		} else {
			error << compose(_("unknown track height name \"%1\" in XML GUI information"), prop->value()) << endmsg;
			set_height (Normal);
		}
	} else {
		set_height (Normal);
	}
}

void
AudioTimeAxisView::set_height (TrackHeight h)
{
	bool height_changed = (h != (TrackHeight)height);
	
	TimeAxisView::set_height (h);

	ensure_xml_node ();
	view->set_height ((double) height);

	gain_line->set_height (height);
	pan_line->set_height (height);

	for (vector<RedirectAutomationLine*>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {
		(*i)->set_height (height);
	}

	if (GTK_OBJECT_FLAGS(GTK_OBJECT(selection_group)) & GTK_CANVAS_ITEM_VISIBLE) {
		/* resize the selection rect */
		show_selection (editor.get_selection().time);
	}

	xml_node->remove_property ("track_height");

	switch (height) {
	case Largest:
		xml_node->add_property ("track_height", "largest");
		name_label.hide ();
		expansion_box.hide ();
		name_entry.show ();
		loop_button.show ();
		edit_group_button.show ();
		history_button.show ();
		other_button_hbox.show_all ();
		button_table.show_all ();
		break;
	case Large:
		xml_node->add_property ("track_height", "large");
		name_label.hide ();
		expansion_box.hide ();
		name_entry.show ();
		loop_button.show ();
		edit_group_button.show ();
		history_button.show ();
		other_button_hbox.show_all ();
		button_table.show_all ();
		break;
	case Larger:
		xml_node->add_property ("track_height", "larger");
		name_label.hide ();
		expansion_box.hide ();
		name_entry.show ();
		loop_button.show ();
		edit_group_button.show ();
		history_button.show ();
		other_button_hbox.show_all ();
		button_table.show_all ();
		break;
	case Normal:
		xml_node->add_property ("track_height", "normal");
		name_label.hide ();
		expansion_box.hide ();
		name_entry.show ();
		loop_button.show ();
		edit_group_button.show ();
		history_button.show ();
		other_button_hbox.show_all ();
		button_table.show_all ();
		break;
	case Smaller:
		xml_node->add_property ("track_height", "smaller");
		name_label.hide ();
		name_entry.show ();
		expansion_box.hide ();
		button_table.show_all ();
		loop_button.hide ();
		edit_group_button.hide ();
		history_button.hide ();
		other_button_hbox.hide_all ();
		break;
	case Small:
		xml_node->add_property ("track_height", "small");
		name_label.set_text (_route.name());
		name_label.show ();
		expansion_box.show ();
		name_entry.hide ();
		button_table.hide_all ();
		other_button_hbox.hide_all ();
		break;
	}

	if (height_changed) {
		/* only emit the signal if the height really changed */
		EMIT_SIGNAL _route.gui_changed ("track_height", (void *) 0);
	}
}

void
AudioTimeAxisView::select_track_color ()
{
	if (RouteUI::choose_color ()) {

		if (view) {
			view->apply_color (_color, StreamView::RegionColor);
		}
	}
}

/* Automation control lines */

void
AudioTimeAxisView::automation_control_point_changed (AutomationType type)
{
	Gtkmmext::UI::instance()->call_slot (bind (slot (*this, &AudioTimeAxisView::update_automation_view), type));
}

void
AudioTimeAxisView::update_automation_view (AutomationType type)
{
	switch (type) {
	case GainAutomation:
		gain_line->reset ();
		break;
	case PanAutomation:
		pan_line->reset ();
		break;
	case PluginAutomation:
		reset_redirect_automation_curves ();
		break;
	case SoloAutomation:
		break;
	case MuteAutomation:
		break;
	};

	
}

void
AudioTimeAxisView::reset_redirect_automation_curves ()
{
	for (vector<RedirectAutomationLine*>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {
		(*i)->reset();
	}
}

void
AudioTimeAxisView::reset_samples_per_unit ()
{
	double speed = 1.0;

	if (get_diskstream() != 0) {
		speed = get_diskstream()->speed();
	}
	
	view->set_samples_per_unit (editor.get_current_zoom() * speed);
	gain_line->reset ();
	pan_line->reset ();
	reset_redirect_automation_curves ();
}

void
AudioTimeAxisView::build_display_menu ()
{
	using namespace Menu_Helpers;

	/* get the size menu ready */

	build_size_menu ();

	/* prepare it */

	TimeAxisView::build_display_menu ();

	/* now fill it with our stuff */

	MenuList& items = display_menu->items();

	items.push_back (MenuElem (_("Height"), *size_menu));
	items.push_back (MenuElem (_("Color"), slot (*this, &AudioTimeAxisView::select_track_color)));

	automation_action_menu = manage (new Menu);
	MenuList& automation_items = automation_action_menu->items();

	automation_clear_menu = manage (new Menu);
	automation_clear_menu->map_event.connect (bind (slot (*this, &AudioTimeAxisView::handle_automation_menu_map_event), automation_clear_menu));
	automation_hide_menu = manage (new Menu);
	automation_hide_menu->map_event.connect (bind (slot (*this, &AudioTimeAxisView::handle_automation_menu_map_event), automation_hide_menu));
	automation_show_menu = manage (new Menu);
	automation_show_menu->map_event.connect (bind (slot (*this, &AudioTimeAxisView::handle_automation_menu_map_event), automation_show_menu));

	automation_items.push_back (MenuElem (_("Clear"), *automation_clear_menu));
	automation_items.push_back (MenuElem (_("Hide"), *automation_hide_menu));
	automation_items.push_back (MenuElem (_("Show"), *automation_show_menu));

	items.push_back (MenuElem (_("Automation"), *automation_action_menu));
			 
	Menu *waveform_menu = manage(new Menu);
	MenuList& waveform_items = waveform_menu->items();

	waveform_items.push_back (CheckMenuElem (_("Show waveforms"), slot (*this, &AudioTimeAxisView::toggle_waveforms)));
	waveform_item = static_cast<CheckMenuItem *> (waveform_items.back());
	ignore_toggle = true;
	waveform_item->set_active (editor.show_waveforms());
	ignore_toggle = false;

	RadioMenuItem::Group group;

	waveform_items.push_back (RadioMenuElem (group, _("Traditional"), bind (slot (*this, &AudioTimeAxisView::set_waveform_shape), Traditional)));
	traditional_item = static_cast<RadioMenuItem *> (waveform_items.back());

	waveform_items.push_back (RadioMenuElem (group, _("Rectified"), bind (slot (*this, &AudioTimeAxisView::set_waveform_shape), Rectified)));
	rectified_item = static_cast<RadioMenuItem *> (waveform_items.back());

	items.push_back (MenuElem (_("Waveform"), *waveform_menu));

	items.push_back (SeparatorElem());
	items.push_back (MenuElem (_("Remove"), slot (*this, &RouteUI::remove_this_route)));
}

gint
AudioTimeAxisView::handle_display_menu_map_event (GdkEventAny *ev)
{
	build_subplugin_menu (ev);
	return FALSE;
}

gint
AudioTimeAxisView::handle_automation_menu_map_event (GdkEventAny *ev, Gtk::Menu* menu)
{
	using namespace Menu_Helpers;

	MenuList& items = menu->items();

	items.clear();

	void (AutomationLine::*action)(void);

	if (menu == automation_clear_menu) {
		action = &AutomationLine::clear;
	} else if (menu == automation_hide_menu) {
		action = &AutomationLine::hide;
	} else if (menu == automation_show_menu) {
		action = &AutomationLine::show;
	} else {
		return FALSE;
	}

	items.push_back (MenuElem ("Gain", slot (*gain_line, action)));
	items.push_back (MenuElem ("Pan", slot (*pan_line, action)));

	for (vector<RedirectAutomationLine *>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {
		string name;
		name = (*i)->redirect().name();
		name += ':';
		name += (*i)->name();

		items.push_back (MenuElem (name, slot ((*i), action)));
	}

	return FALSE;
}

void
AudioTimeAxisView::build_playlist_menu (Gtk::Menu * menu)
{
	using namespace Menu_Helpers;

	if (!menu) return;
	
	if (is_audio_track()) {
		MenuList& playlist_items = menu->items();
		playlist_items.clear();
		
		playlist_items.push_back (MenuElem (_("New"), slot (*this, &AudioTimeAxisView::use_new_playlist)));
		playlist_items.push_back (MenuElem (_("New Copy"), slot(*this, &AudioTimeAxisView::use_copy_playlist)));
		playlist_items.push_back (SeparatorElem());
		playlist_items.push_back (MenuElem (_("Clear Current"), slot (*this, &AudioTimeAxisView::clear_playlist)));


		// since choose is so common, lets put the playlists created on this track in the main menu
		playlist_items.push_back (SeparatorElem());
		
		playlist_menu_toplist.clear();
		playlist_menu_botlist.clear();
		
		_session.foreach_playlist (this, &AudioTimeAxisView::add_playlist_to_playlist_menu);

		for (vector<Playlist *>::iterator pl = playlist_menu_toplist.begin(); pl != playlist_menu_toplist.end(); ++pl) {
			playlist_items.push_back (CheckMenuElem ((*pl)->name(), bind (slot (*this, &AudioTimeAxisView::playlist_selected), *pl)));

			CheckMenuItem * mitem = (CheckMenuItem *) playlist_items.back(); 
			if (get_diskstream()->playlist() == *pl) {
				mitem->set_active(true);
			}
		}

		// make a submenu that contains all the others
		playlist_menu = new Menu;
		MenuList & sublist = playlist_menu->items();

		Playlist * otherpl = 0;
		
		for (vector<Playlist *>::iterator pl = playlist_menu_botlist.begin(); pl != playlist_menu_botlist.end(); ++pl) {
			sublist.push_back (CheckMenuElem ((*pl)->name(), bind (slot (*this, &AudioTimeAxisView::playlist_selected), *pl)));

			CheckMenuItem * mitem = (CheckMenuItem *) sublist.back(); 
			if (get_diskstream()->playlist() == *pl) {
				otherpl = *pl;
				mitem->set_active(true);
			}
		}

		if (otherpl) {
			// place current selected other playlist in main menu for clarity
			playlist_items.push_back (SeparatorElem());
			playlist_items.push_back (CheckMenuElem (otherpl->name(), bind (slot (*this, &AudioTimeAxisView::playlist_selected), otherpl)));
			CheckMenuItem * mitem = (CheckMenuItem *) playlist_items.back(); 
			mitem->set_active (true);
		}
		
		playlist_items.push_back (SeparatorElem());
		playlist_items.push_back (MenuElem(_("Use Other"), *playlist_menu));
	}

}


void
AudioTimeAxisView::add_playlist_to_playlist_menu (Playlist *pl)
{
	using namespace Menu_Helpers;

	if (is_audio_track() && pl->get_orig_diskstream_id() == get_diskstream()->id()) {
		playlist_menu_toplist.push_back (pl);
	}
	else {
		playlist_menu_botlist.push_back (pl);
	}
}

void
AudioTimeAxisView::playlist_selected (Playlist *pl)
{
	DiskStream *ds;

	if ((ds = get_diskstream()) != 0) {
		ds->use_playlist (pl);
	}
}

void
AudioTimeAxisView::use_copy_playlist ()
{
	DiskStream *ds;

	if ((ds = get_diskstream()) != 0) {
		ds->use_copy_playlist ();
	}
}

void
AudioTimeAxisView::use_new_playlist ()
{
	DiskStream *ds;

	if ((ds = get_diskstream()) != 0) {
		ds->use_new_playlist ();
	}
}	

void
AudioTimeAxisView::clear_playlist ()
{
	Playlist *pl;
	DiskStream *ds;
	
	if ((ds = get_diskstream()) != 0) {
		if ((pl = ds->playlist()) != 0) {
			editor.clear_playlist (*pl);
		}
	}
}

void
AudioTimeAxisView::toggle_waveforms ()
{
	if (view && waveform_item && !ignore_toggle) {
		view->set_show_waveforms (waveform_item->is_active());
	}
}

void
AudioTimeAxisView::set_show_waveforms (bool yn)
{
	if (waveform_item) {
		waveform_item->set_active (yn);
	} else {
		view->set_show_waveforms (yn);
	}
}

void
AudioTimeAxisView::set_waveform_shape (WaveformShape shape)
{
	if (view) {
		view->set_waveform_shape (shape);
	}
}

void
AudioTimeAxisView::speed_changed ()
{
	Gtkmmext::UI::instance()->call_slot (slot (*this, &AudioTimeAxisView::reset_samples_per_unit));
}

void
AudioTimeAxisView::diskstream_changed (void *src)
{
	Gtkmmext::UI::instance()->call_slot (slot (*this, &AudioTimeAxisView::update_diskstream_display));
}	

void
AudioTimeAxisView::update_diskstream_display ()
{
	DiskStream *ds;

	if ((ds = get_diskstream()) != 0) {
		set_playlist (ds->playlist ());
		rec_enable_button.set_sensitive (true);
		playlist_button.set_sensitive (true);
		ds->LoopSet.connect (slot (*this, &AudioTimeAxisView::loop_set));
	} else {
		playlist_button.set_sensitive (false);
		rec_enable_button.set_sensitive (false);
	}
}	

void
AudioTimeAxisView::add_redirect_to_subplugin_menu (Redirect* r)
{
	using namespace Menu_Helpers;

	Menu* lines = manage (new Menu);
	MenuList& items = lines->items();
	
	/* create a submenu with each currently-existing parameter curve on it */

	for (vector<RedirectAutomationLine *>::iterator i = redirect_automation_curves.begin(); i != redirect_automation_curves.end(); ++i) {

		if (&(*i)->redirect() == r) {

			items.push_back (CheckMenuElem ((*i)->name()));

			CheckMenuItem *item = dynamic_cast<CheckMenuItem*>(items.back());
			
			if ((*i)->visible()) {
				item->set_active (false);
			} else {
				item->set_active (true);
			}
		}
	}

	/* add the element */

	subplugin_menu.items().push_back (MenuElem (r->name(), *lines));
}

gint
AudioTimeAxisView::build_subplugin_menu (GdkEventAny* ev)
{
	Menu_Helpers::MenuList& subplugin_items = subplugin_menu.items();

	subplugin_items.clear ();
	_route.foreach_redirect (this, &AudioTimeAxisView::add_redirect_to_subplugin_menu);

	return FALSE;
}

void
AudioTimeAxisView::add_gain_automation_point_event (GtkCanvasItem *item, GdkEvent *ev, jack_nframes_t frame, double y)
{
	double x = 0;

	gtk_canvas_item_w2i (canvas_display, &x, &y);

	/* compute vertical fractional position */

	y = 1.0 - (y / height);

	/* map using gain line */

	gain_line->view_to_model_y (y);

	_session.begin_reversible_command ("add gain CP");
	_session.add_undo (_route.get_memento());
	_route.add_automation_event (GainAutomation, frame, y, true);
	_session.add_redo_no_execute (_route.get_memento());
	_session.commit_reversible_command ();
}

void
AudioTimeAxisView::order_selection_trims (GtkCanvasItem *item, bool put_start_on_top)
{
	/* find the selection rect this is for. we have the item corresponding to one
	   of the trim handles.
	 */

	for (list<SelectionRect*>::iterator i = used_selection_rects.begin(); i != used_selection_rects.end(); ++i) {
		if ((*i)->start_trim == item || (*i)->end_trim == item) {

			/* make one trim handle be "above" the other so that if they overlap,
			   the top one is the one last used.
			*/

			gtk_canvas_item_raise_to_top ((*i)->rect);
			gtk_canvas_item_raise_to_top (put_start_on_top ? (*i)->start_trim : (*i)->end_trim);
			gtk_canvas_item_raise_to_top (put_start_on_top ? (*i)->end_trim : (*i)->start_trim);

			break;
		}
	}
}

void
AudioTimeAxisView::set_selected_regionviews (RegionSelection& regions)
{
	if (view) {
		view->set_selected_regionviews (regions);
	}
}

void
AudioTimeAxisView::set_samples_per_unit (double spu)
{
	double speed = 1.0;

	if (get_diskstream() != 0) {
		speed = get_diskstream()->speed();
	}
	
	if (view) {
		view->set_samples_per_unit (spu * speed);
	}
}

RouteGroup*
AudioTimeAxisView::edit_group() const
{
	return _route.edit_group();
}

string
AudioTimeAxisView::name() const
{
	return _route.name();
}

Playlist *
AudioTimeAxisView::playlist () const 
{
	DiskStream *ds;

	if ((ds = get_diskstream()) != 0) {
		return ds->playlist(); 
	} else {
		return 0; 
	}
}

gint 
AudioTimeAxisView::name_entry_button_press_handler (GdkEventButton *ev)
{
	if (ev->button == 3) {
		return stop_signal (name_entry, "button_press_event");
	}
	return FALSE;
}

gint 
AudioTimeAxisView::name_entry_button_release_handler (GdkEventButton *ev)
{
	return FALSE;
}

gint
AudioTimeAxisView::name_entry_focus_out_handler (GdkEventFocus* ev)
{
	name_entry_changed ();
	return TRUE;
}

gint
AudioTimeAxisView::name_entry_key_release_handler (GdkEventKey* ev)
{
	switch (ev->keyval) {
	case GDK_Tab:
	case GDK_Up:
	case GDK_Down:
		name_entry_changed ();
		return TRUE;

	default:
		return FALSE;
	}
}

void
AudioTimeAxisView::name_entry_activated ()
{
	/* this should drop focus from the entry,
	   and cause a call to name_entry_changed()
	*/
	controls_ebox.grab_focus();
}

void
AudioTimeAxisView::name_entry_changed ()
{
	string x;

	ARDOUR_UI::generic_focus_out_event (0);

	x = name_entry.get_text ();
	
	if (x == _route.name()) {
		return;
	}

	if (x.length() == 0) {
		name_entry.set_text (_route.name());
		return;
	}

	strip_whitespace_edges(x);

	if (_session.route_name_unique (x)) {
		_route.set_name (x, this);
	} else {
		ARDOUR_UI::instance()->popup_error (_("a track already exists with that name"));
		name_entry.set_text (_route.name());
	}
}

void
AudioTimeAxisView::visual_click ()
{
	popup_display_menu (0);
}

void
AudioTimeAxisView::hide_click ()
{
	editor.unselect_strip_in_display (*this);
}

Region*
AudioTimeAxisView::find_next_region (jack_nframes_t pos, Region::Point point, int dir)
{
	DiskStream *stream;
	Playlist *playlist;

	if ((stream = get_diskstream()) != 0 && (playlist = stream->playlist()) != 0) {
		return playlist->find_next_region (pos, point, dir);
	}

	return 0;
}
