/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: editor_rulers.cc,v 1.22 2003/10/21 04:33:04 essej Exp $
*/

#include <cstdio> // for sprintf, grrr 
#include <cmath>

#include <string>

#include <ardour/tempo.h>

#include "editor.h"
#include "gtk-custom-hruler.h"

using namespace SigC;
using namespace ARDOUR;
using namespace Gtk;

Editor *Editor::ruler_editor;

/* the order here must match the "metric" enums in editor.h */

GtkCustomMetric Editor::ruler_metrics[4] = {
	{1, Editor::_metric_get_smpte },
	{1, Editor::_metric_get_bbt },
	{1, Editor::_metric_get_frames },
	{1, Editor::_metric_get_minsec }
};

void
Editor::initialize_rulers ()
{
	ruler_editor = this;

	_smpte_ruler = gtk_custom_hruler_new ();
	smpte_ruler = wrap (_smpte_ruler);
	smpte_ruler->set_name ("SMPTERuler");
	smpte_ruler->set_usize (-1, (int)timebar_height);
	gtk_custom_ruler_set_metric (GTK_CUSTOM_RULER(_smpte_ruler), &ruler_metrics[ruler_metric_smpte]);

	_bbt_ruler = gtk_custom_hruler_new ();
	bbt_ruler = wrap (_bbt_ruler);
	bbt_ruler->set_name ("BBTRuler");
	bbt_ruler->set_usize (-1, (int)timebar_height);
	gtk_custom_ruler_set_metric (GTK_CUSTOM_RULER(_bbt_ruler), &ruler_metrics[ruler_metric_bbt]);

	_frames_ruler = gtk_custom_hruler_new ();
	frames_ruler = wrap (_frames_ruler);
	frames_ruler->set_name ("FramesRuler");
	frames_ruler->set_usize (-1, (int)timebar_height);
	gtk_custom_ruler_set_metric (GTK_CUSTOM_RULER(_frames_ruler), &ruler_metrics[ruler_metric_frames]);

	_minsec_ruler = gtk_custom_hruler_new ();
	minsec_ruler = wrap (_minsec_ruler);
	frames_ruler->set_name ("MinSecRuler");
	minsec_ruler->set_usize (-1, (int)timebar_height);
	gtk_custom_ruler_set_metric (GTK_CUSTOM_RULER(_minsec_ruler), &ruler_metrics[ruler_metric_minsec]);

	smpte_ruler->set_events (GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);
	bbt_ruler->set_events (GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);
	frames_ruler->set_events (GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);
	minsec_ruler->set_events (GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);

	smpte_ruler->button_release_event.connect (slot (*this, &Editor::ruler_button_release));
	bbt_ruler->button_release_event.connect (slot (*this, &Editor::ruler_button_release));
	frames_ruler->button_release_event.connect (slot (*this, &Editor::ruler_button_release));
	minsec_ruler->button_release_event.connect (slot (*this, &Editor::ruler_button_release));

	smpte_ruler->button_press_event.connect (slot (*this, &Editor::ruler_button_press));
	bbt_ruler->button_press_event.connect (slot (*this, &Editor::ruler_button_press));
	frames_ruler->button_press_event.connect (slot (*this, &Editor::ruler_button_press));
	minsec_ruler->button_press_event.connect (slot (*this, &Editor::ruler_button_press));
	
	smpte_ruler->motion_notify_event.connect (slot (*this, &Editor::ruler_mouse_motion));
	bbt_ruler->motion_notify_event.connect (slot (*this, &Editor::ruler_mouse_motion));
	frames_ruler->motion_notify_event.connect (slot (*this, &Editor::ruler_mouse_motion));
	minsec_ruler->motion_notify_event.connect (slot (*this, &Editor::ruler_mouse_motion));
	
	visible_timebars = 7; /* 4 here, 3 in time_canvas */
	ruler_pressed_button = 0;
}


gint
Editor::ruler_button_press (GdkEventButton* ev)
{
	if (session == 0) {
		return FALSE;
	}

	ruler_pressed_button = ev->button;
	
	return TRUE;
}

gint
Editor::ruler_button_release (GdkEventButton* ev)
{
	ruler_pressed_button = 0;
	
	if (session == 0) {
		return FALSE;
	}

	hide_verbose_canvas_cursor();
	stop_canvas_autoscroll();
	
	jack_nframes_t where = leftmost_frame + pixel_to_frame (ev->x);

	snap_to (where);

	switch (ev->button) {
	case 1:
		/* transport playhead */
		session->request_locate (where);
		break;

	case 2:
		/* edit cursor */
		edit_cursor->set_position (where);
		break;

	default:
		break;
	}

	return TRUE;
}

gint
Editor::ruler_mouse_motion (GdkEventMotion* ev)
{
	if (session == 0 || !ruler_pressed_button) {
		return FALSE;
	}

	double wcx=0,wcy=0;
	double cx=0,cy=0;

	
	gtk_canvas_window_to_world (GTK_CANVAS(track_gtk_canvas), ev->x, ev->y, &wcx, &wcy);
	gtk_canvas_w2c_d (GTK_CANVAS(track_gtk_canvas), wcx, wcy, &cx, &cy);
	
	jack_nframes_t where = leftmost_frame + pixel_to_frame (ev->x);

	/// ripped from maybe_autoscroll
	jack_nframes_t one_page = (jack_nframes_t) rint (canvas_width * frames_per_unit);
	jack_nframes_t rightmost_frame = leftmost_frame + one_page;

	jack_nframes_t frame = pixel_to_frame (cx);

	if (autoscroll_timeout_tag < 0) {
		if (frame > rightmost_frame) {
			if (rightmost_frame < max_frames) {
				start_canvas_autoscroll (1);
			}
		} else if (frame < leftmost_frame) {
			if (leftmost_frame > 0) {
				start_canvas_autoscroll (-1);
			}
		} 
	} else {
		if (frame >= leftmost_frame && frame < rightmost_frame) {
			stop_canvas_autoscroll ();
		}
	}
	//////	
	
	snap_to (where);

	Cursor* cursor = 0;
	
	switch (ruler_pressed_button) {
	case 1:
		/* transport playhead */
		cursor = playhead_cursor;
		break;

	case 2:
		/* edit cursor */
		cursor = edit_cursor;
		break;

	default:
		break;
	}

	if (cursor)
	{
		cursor->set_position (where);
		
		if (cursor == edit_cursor) {
			edit_cursor_clock.set (where);
		}
		
		show_verbose_time_cursor (where, 10, cx, 0);
	}
	
	return TRUE;
}


void
Editor::update_fixed_rulers ()
{
	jack_nframes_t rightmost_frame;

	if (session == 0) {
		return;
	}

	/* XXX Note the potential loss of accuracy here as we convert from
	   an unsigned long (or larger) to a float ... what to do ?
	*/

	jack_nframes_t page = (jack_nframes_t) floor (canvas_width * frames_per_unit);

	ruler_metrics[ruler_metric_smpte].units_per_pixel = frames_per_unit;
	ruler_metrics[ruler_metric_frames].units_per_pixel = frames_per_unit;
	ruler_metrics[ruler_metric_minsec].units_per_pixel = frames_per_unit;

	rightmost_frame = leftmost_frame + page;

	/* these force a redraw, which in turn will force execution of the metric callbacks
	   to compute the relevant ticks to display.
	*/

	gtk_custom_ruler_set_range (GTK_CUSTOM_RULER(_smpte_ruler), leftmost_frame, rightmost_frame,
				    leftmost_frame, session->current_end_frame());
	gtk_custom_ruler_set_range (GTK_CUSTOM_RULER(_frames_ruler), leftmost_frame, rightmost_frame,
				    leftmost_frame, session->current_end_frame());
	gtk_custom_ruler_set_range (GTK_CUSTOM_RULER(_minsec_ruler), leftmost_frame, rightmost_frame,
				    leftmost_frame, session->current_end_frame());
}		

void
Editor::update_tempo_based_rulers ()
{
	if (session == 0) {
		return;
	}

	/* XXX Note the potential loss of accuracy here as we convert from
	   an unsigned long (or larger) to a float ... what to do ?
	*/

	jack_nframes_t page = (jack_nframes_t) floor (canvas_width * frames_per_unit);
	ruler_metrics[ruler_metric_bbt].units_per_pixel = frames_per_unit;
	gtk_custom_ruler_set_range (GTK_CUSTOM_RULER(_bbt_ruler), leftmost_frame, leftmost_frame+page, 
				    leftmost_frame, session->current_end_frame());
}

/* Mark generation */

gint
Editor::_metric_get_smpte (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	return ruler_editor->metric_get_smpte (marks, lower, upper, maxchars);
}

gint
Editor::_metric_get_bbt (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	return ruler_editor->metric_get_bbt (marks, lower, upper, maxchars);
}

gint
Editor::_metric_get_frames (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	return ruler_editor->metric_get_frames (marks, lower, upper, maxchars);
}

gint
Editor::_metric_get_minsec (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	return ruler_editor->metric_get_minsec (marks, lower, upper, maxchars);
}

gint
Editor::metric_get_smpte (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	jack_nframes_t range;
	jack_nframes_t fr;
	jack_nframes_t mark_interval;
	jack_nframes_t pos;
	SMPTE_Time smpte;
	gchar buf[16];
	gint nmarks;
	gint n;

	if (session == 0) {
		return 0;
	}

	fr = session->frame_rate();
	range = upper - lower;

	if (range < session->frames_per_smpte_frame()) { 
		mark_interval = session->frames_per_smpte_frame(); /* show SMPTE frames */
	} else if (range <= (jack_nframes_t) floor(fr * 0.1)) { /* 0-0.1 second */
		mark_interval = (jack_nframes_t) (fr * 0.025); /* show 1/40 seconds */
	} else if (range <= fr/2) { /* 0-0.5 second */
		mark_interval = (jack_nframes_t) (fr * 0.05);  /* show 1/20 seconds */
	} else if (range <= fr) { /* 0-1 second */
		mark_interval = (jack_nframes_t) (fr * 0.1);  /* show 1/10 seconds */
	} else if (range <= 2 * fr) { /* 1-2 seconds */
		mark_interval = (jack_nframes_t) floor (fr * 0.5); /* show 1/2 seconds */
	} else if (range <= 8 * fr) { /* 2-5 seconds */
		mark_interval =  2 * fr; /* show 2 seconds */
	} else if (range <= 16 * fr) { /* 8-16 seconds */
		mark_interval =  4 * fr; /* show 4 seconds */
	} else if (range <= 30 * fr) { /* 10-30 seconds */
		mark_interval = 10 * fr; /* show 10 seconds */
	} else if (range <= 60 * fr) { /* 30-60 seconds */
		mark_interval = 15 * fr; /* show 15 seconds */
	} else if (range <= 2 * 60 * fr) { /* 1-2 minutes */
		mark_interval = 30 * fr; /* show 30 seconds */
	} else if (range <= 10 * 60 * fr) { /* 2-10 minutes */
		mark_interval = 60 * fr; /* show 1 minutes */
	} else if (range <= 15 * 60 * fr) { /* 10-15 minutes */
		mark_interval = 2 * 60 * fr; /* show 2 minutes */
	} else if (range <= 30 * 60 * fr) { /* 10-30 minutes */
		mark_interval = 5 * 60 * fr; /* show 5 minutes */
	} else if (range <= 60 * 60 * fr) { /* 30 minutes - 1hr */
		mark_interval = 10 * 60 * fr; /* show 10 minutes */
	} else if (range <= 4 * 60 * 60 * fr) { /* 1 - 4 hrs*/
		mark_interval = 30 * 60 * fr; /* show 30 minutes */
	} else if (range <= 8 * 60 * 60 * fr) { /* 4 - 8 hrs*/
		mark_interval = 60 * 60 * fr; /* show hrs */
	} else if (range <= 16 * 60 * 60 * fr) { /* 8-16 hrs*/
		mark_interval = (jack_nframes_t) floor (1.5 * 60 * 60 * fr); /* show 1 1/2hrs XXX don't like this */
	} else if (range <= 16 * 60 * 60 * fr) { /* 16-24 hrs*/
		mark_interval = 2 * 60 * 60 * fr; /* show 2 hrs */
	} else {

		/* not possible if jack_nframes_t is a 32 bit quantity */

		mark_interval = 4 * 60 * 60 * fr; /* show 4 hrs */
	}

	nmarks = 1+ (range / mark_interval);
	*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
	pos = ((lower + (mark_interval/2))/mark_interval) * mark_interval;
	pos += session->smpte_offset(0);

	for (n = 0; n < nmarks; pos += mark_interval, ++n) {
		session->smpte_time (pos, smpte);
		if (smpte.negative) {
			snprintf (buf, sizeof(buf), "-%02ld:%02ld:%02ld:%02ld", smpte.hours, smpte.minutes, smpte.seconds, smpte.frames);
		} else {
			snprintf (buf, sizeof(buf), "%02ld:%02ld:%02ld:%02ld", smpte.hours, smpte.minutes, smpte.seconds, smpte.frames);
		}
		(*marks)[n].label = g_strdup (buf);
		(*marks)[n].position = pos;
		(*marks)[n].style = GtkCustomRulerMarkMajor;
	}

	return nmarks;
}

gint
Editor::metric_get_bbt (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
        if (session == 0) {
                return 0;
        }

        TempoMap::BBTPointList::iterator i;
	TempoMap::BBTPointList *zoomed_bbt_points;
        unsigned long beats = 0;
        unsigned long bars = 0;
	unsigned long tick = 0;
	unsigned long skip;
	unsigned long t;
        unsigned long zoomed_beats = 0;
        unsigned long zoomed_bars = 0;
        unsigned long desirable_marks;
	gint nmarks;
        char buf[64];
        gint n;
	jack_nframes_t pos;
        jack_nframes_t frame_one_beats_worth;
        jack_nframes_t frame_skip;
        bool bar_helper_on = false;

	if ((desirable_marks = maxchars / 6) == 0) {
                return 0;
        }

/*      align the tick marks to whatever we're snapping to... */
                                                                                                             
        if (snap_type == SnapToAThirdBeat) {
                bbt_beat_subdivision = 3;
        } else if (snap_type == SnapToAQuarterBeat) {
                bbt_beat_subdivision = 4;
        }

       	skip = (unsigned long) (Meter::ticks_per_beat / bbt_beat_subdivision);

	frame_one_beats_worth = (jack_nframes_t) ::floor ((double) session->frame_rate() *  60 / 20 ); //one beat @ 20 bpm
	zoomed_bbt_points = session->tempo_map().get_points((lower >= frame_one_beats_worth) ? lower - frame_one_beats_worth : 0, upper + frame_one_beats_worth);

	if (zoomed_bbt_points == 0 || zoomed_bbt_points->empty()) {
		return 0;
	}

	for (i = current_bbt_points->begin(); i != current_bbt_points->end(); i++) {
        	if ((*i).type == TempoMap::Beat) {
			beats++;
		} else if ((*i).type == TempoMap::Bar) {
			bars++;
		}
	}

	if (bars == 0) {
		bar_helper_on = true;
	}

	for (i = zoomed_bbt_points->begin(); i != zoomed_bbt_points->end(); i++) {
               	if ((*i).type == TempoMap::Beat) {
                       	zoomed_beats++;
               	} else if ((*i).type == TempoMap::Bar) {
                       	zoomed_bars++;
               	}
       	}

	if (desirable_marks > (beats / 4)) {

		if (desirable_marks > beats) {
               		nmarks = (zoomed_beats * bbt_beat_subdivision) + 1;
		} else {
			nmarks = zoomed_beats + 1;
		}

		*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
		for (n = 1, i = zoomed_bbt_points->begin(); i != zoomed_bbt_points->end() && n < nmarks; ++i) {
			if ((*i).frame <= lower) {
				if (bar_helper_on) {
					snprintf (buf, sizeof(buf), "<%lu|%lu", (*i).bar, (*i).beat);
				} else {
					snprintf (buf, sizeof(buf), " ");
				}
				(*marks)[0].label = g_strdup (buf);
				(*marks)[0].position = lower;
				(*marks)[0].style = GtkCustomRulerMarkMajor;
			}
			if ((*i).type == TempoMap::Bar)  {
				tick = 0;
				(((*i).frame < (lower + (30 * Editor::get_current_zoom ()))) && (bar_helper_on)) ?
					snprintf (buf, sizeof(buf), " ") : snprintf (buf, sizeof(buf), "%lu", (*i).bar);
				(*marks)[n].label = g_strdup (buf);
				(*marks)[n].position = (*i).frame;
				(*marks)[n].style = GtkCustomRulerMarkMajor;
				n++;
			} else if (((*i).type == TempoMap::Beat) && ((*i).beat > 1)) {
				tick = 0;
				((((*i).frame < (lower + (30 * Editor::get_current_zoom ()))) && (bar_helper_on)) || (desirable_marks < beats)) ?
					snprintf (buf, sizeof(buf), " ") : snprintf (buf, sizeof(buf), "%lu", (*i).beat);
				if (desirable_marks < beats) {
					(*marks)[n].style = GtkCustomRulerMarkMicro;
				} else {
					(*marks)[n].style = GtkCustomRulerMarkMinor;
				}
				(*marks)[n].label =  g_strdup (buf);
				(*marks)[n].position = (*i).frame;
				n++;
			}

			/* Add the tick marks */


			if (desirable_marks > beats) {
				frame_skip = (jack_nframes_t) floor ((session->frame_rate() *  60) / (bbt_beat_subdivision * (*i).tempo->beats_per_minute()));
				pos = (*i).frame + frame_skip;
				tick += skip;
				for (t = 0; tick < Meter::ticks_per_beat && n < nmarks ; pos += frame_skip, tick += skip, ++t) {
					if (Editor::get_current_zoom () > 32) {
						snprintf (buf, sizeof(buf), " ");
					} else  if (pos < (lower + (30 * Editor::get_current_zoom ()))) {
						snprintf (buf, sizeof(buf), " ");
					} else {
						snprintf (buf, sizeof(buf), "%lu", tick);
					}
					(*marks)[n].label = g_strdup (buf);
					(*marks)[n].position = pos;
					(*marks)[n].style = GtkCustomRulerMarkMicro;
					n++;
				}
			}
		}
		delete zoomed_bbt_points;
		return nmarks;

       } else {

		if (desirable_marks < (unsigned long) (zoomed_bars / 256)) {
        		nmarks = 1;
			*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
			snprintf (buf, sizeof(buf), "too many bars... (currently %lu)", zoomed_bars );
        		(*marks)[0].style = GtkCustomRulerMarkMajor;
        		(*marks)[0].label = g_strdup (buf);
			(*marks)[0].position = lower;
		} else if (desirable_marks < (unsigned long) (nmarks = (gint) (zoomed_bars / 64))) {
			*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
			for (n = 0, i = zoomed_bbt_points->begin(); i != zoomed_bbt_points->end() && n < nmarks; i++) {
				if ((*i).type == TempoMap::Bar)  {
					if ((*i).bar % 64 == 1) {
						if ((*i).bar % 256 == 1) {
							snprintf (buf, sizeof(buf), "%lu", (*i).bar);
							(*marks)[n].style = GtkCustomRulerMarkMajor;
						} else {
							snprintf (buf, sizeof(buf), " ");
							if ((*i).bar % 256 == 129)  {
								(*marks)[n].style = GtkCustomRulerMarkMinor;
							} else {
								(*marks)[n].style = GtkCustomRulerMarkMicro;
							}
						}
						(*marks)[n].label = g_strdup (buf);
						(*marks)[n].position = (*i).frame;
						n++;
					}
				}
			}
		} else if (desirable_marks < (unsigned long) (nmarks = (gint)(zoomed_bars / 16))) {
			*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
			for (n = 0, i = zoomed_bbt_points->begin(); i != zoomed_bbt_points->end() && n < nmarks; i++) {
				if ((*i).type == TempoMap::Bar)  {
					if ((*i).bar % 16 == 1) {
						if ((*i).bar % 64 == 1) {
							snprintf (buf, sizeof(buf), "%lu", (*i).bar);
							(*marks)[n].style = GtkCustomRulerMarkMajor;
						} else {
							snprintf (buf, sizeof(buf), " ");
							if ((*i).bar % 64 == 33)  {
								(*marks)[n].style = GtkCustomRulerMarkMinor;
							} else {
								(*marks)[n].style = GtkCustomRulerMarkMicro;
							}
						}
						(*marks)[n].label = g_strdup (buf);
						(*marks)[n].position = (*i).frame;
						n++;
					}
				}
			}
		} else if (desirable_marks < (unsigned long) (nmarks = (gint)(zoomed_bars / 4))){
			*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
			for (n = 0, i = zoomed_bbt_points->begin(); i != zoomed_bbt_points->end() && n < nmarks; ++i) {
				if ((*i).type == TempoMap::Bar)  {
					if ((*i).bar % 4 == 1) {
						if ((*i).bar % 16 == 1) {
							snprintf (buf, sizeof(buf), "%lu", (*i).bar);
							(*marks)[n].style = GtkCustomRulerMarkMajor;
						} else {
							snprintf (buf, sizeof(buf), " ");
							if ((*i).bar % 16 == 9)  {
								(*marks)[n].style = GtkCustomRulerMarkMinor;
							} else {
								(*marks)[n].style = GtkCustomRulerMarkMicro;
							}
						}
						(*marks)[n].label = g_strdup (buf);
						(*marks)[n].position = (*i).frame;
						n++;
					}
				}
			}
		} else {
			nmarks = zoomed_bars;
			*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
                	for (n = 0, i = zoomed_bbt_points->begin(); i != zoomed_bbt_points->end() && n < nmarks; i++) {
                        	if ((*i).type == TempoMap::Bar)  {
                                	if ((*i).bar % 4 == 1) {
                                        	snprintf (buf, sizeof(buf), "%lu", (*i).bar);
                                        	(*marks)[n].style = GtkCustomRulerMarkMajor;
                                	} else {
                                        	snprintf (buf, sizeof(buf), " ");
						if ((*i).bar % 4 == 3)  {
							(*marks)[n].style = GtkCustomRulerMarkMinor;
						} else {
                                        		(*marks)[n].style = GtkCustomRulerMarkMicro;
						}
                                	}
                                	(*marks)[n].label = g_strdup (buf);
                                	(*marks)[n].position = (*i).frame;
                                	n++;
                        	}
                	}
        	}
		delete zoomed_bbt_points;
		return nmarks;
	}
}

gint
Editor::metric_get_frames (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	jack_nframes_t mark_interval;
	jack_nframes_t pos;
	gchar buf[16];
	gint nmarks;
	gint n;

	if (session == 0) {
		return 0;
	}

	mark_interval = (upper - lower) / 5;
	nmarks = 5;
	*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
	for (n = 0, pos = lower; n < nmarks; pos += mark_interval, ++n) {
		snprintf (buf, sizeof(buf), "%u", pos);
		(*marks)[n].label = g_strdup (buf);
		(*marks)[n].position = pos;
		(*marks)[n].style = GtkCustomRulerMarkMajor;
	}
	
	return nmarks;
}

static void
sample_to_clock_parts (guint32 sample,
		       guint32 sample_rate, 
		       float smpte_fps,
		       guint32 *hrs_p,
		       guint32 *mins_p,
		       float *secs_p)

{
	jack_nframes_t left;
	guint32 hrs;
	guint32 mins;
	float secs;
	
	left = sample;
	hrs = (int) floor (left / (sample_rate * 60.0f * 60.0f));
	left -= (jack_nframes_t) floor (hrs * sample_rate * 60.0f * 60.0f);
	mins = (int) floor (left / (sample_rate * 60.0f));
	left -= (jack_nframes_t) floor (mins * sample_rate * 60.0f);
	secs = left / (float) sample_rate;

	*secs_p = secs;
	*mins_p = mins;
	*hrs_p = hrs;

	return;
}

gint
Editor::metric_get_minsec (GtkCustomRulerMark **marks, gulong lower, gulong upper, gint maxchars)
{
	jack_nframes_t range;
	jack_nframes_t fr;
	jack_nframes_t mark_interval;
	jack_nframes_t pos;
	guint32 hrs, mins;
	float secs;
	gchar buf[16];
	gint nmarks;
	gint n;

	if (session == 0) {
		return 0;
	}

	fr = session->frame_rate();
	range = upper - lower;

	if (range < (jack_nframes_t) floor (fr * 0.01)) {
		mark_interval = (jack_nframes_t) (fr * 0.01); /* show 1/100 seconds */
	} else if (range <= (jack_nframes_t) floor(fr * 0.1)) { /* 0-0.1 second */
		mark_interval = (jack_nframes_t) (fr * 0.025); /* show 1/40 seconds */
	} else if (range <= fr/2) { /* 0-0.5 second */
		mark_interval = (jack_nframes_t) (fr * 0.05);  /* show 1/20 seconds */
	} else if (range <= fr) { /* 0-1 second */
		mark_interval = (jack_nframes_t) (fr * 0.1);  /* show 1/10 seconds */
	} else if (range <= 2 * fr) { /* 1-2 seconds */
		mark_interval = (jack_nframes_t) floor (fr * 0.5); /* show 1/2 seconds */
	} else if (range <= 8 * fr) { /* 2-5 seconds */
		mark_interval =  2 * fr; /* show 2 seconds */
	} else if (range <= 16 * fr) { /* 8-16 seconds */
		mark_interval =  4 * fr; /* show 4 seconds */
	} else if (range <= 30 * fr) { /* 10-30 seconds */
		mark_interval = 10 * fr; /* show 10 seconds */
	} else if (range <= 60 * fr) { /* 30-60 seconds */
		mark_interval = 15 * fr; /* show 15 seconds */
	} else if (range <= 2 * 60 * fr) { /* 1-2 minutes */
		mark_interval = 30 * fr; /* show 30 seconds */
	} else if (range <= 10 * 60 * fr) { /* 2-10 minutes */
		mark_interval = 60 * fr; /* show 1 minutes */
	} else if (range <= 15 * 60 * fr) { /* 10-15 minutes */
		mark_interval = 2 * 60 * fr; /* show 2 minutes */
	} else if (range <= 30 * 60 * fr) { /* 10-30 minutes */
		mark_interval = 5 * 60 * fr; /* show 5 minutes */
	} else if (range <= 60 * 60 * fr) { /* 30 minutes - 1hr */
		mark_interval = 10 * 60 * fr; /* show 10 minutes */
	} else if (range <= 4 * 60 * 60 * fr) { /* 1 - 4 hrs*/
		mark_interval = 30 * 60 * fr; /* show 30 minutes */
	} else if (range <= 8 * 60 * 60 * fr) { /* 4 - 8 hrs*/
		mark_interval = 60 * 60 * fr; /* show hrs */
	} else if (range <= 16 * 60 * 60 * fr) { /* 8-16 hrs*/
		mark_interval = (jack_nframes_t) floor (1.5 * 60 * 60 * fr); /* show 1 1/2hrs XXX don't like this */
	} else if (range <= 16 * 60 * 60 * fr) { /* 16-24 hrs*/
		mark_interval = 2 * 60 * 60 * fr; /* show 2 hrs */
	} else {

		/* not possible if jack_nframes_t is a 32 bit quantity */

		mark_interval = 4 * 60 * 60 * fr; /* show 4 hrs */
	}

	nmarks = 1 + (range / mark_interval);
	*marks = (GtkCustomRulerMark *) g_malloc (sizeof(GtkCustomRulerMark) * nmarks);
	pos = ((lower + (mark_interval/2))/mark_interval) * mark_interval;
	
	for (n = 0; n < nmarks; pos += mark_interval, ++n) {
		sample_to_clock_parts (pos, fr, session->smpte_frames_per_second, &hrs, &mins, &secs);
		snprintf (buf, sizeof(buf), "%02d:%02d:%.3f", hrs, mins, secs);
		(*marks)[n].label = g_strdup (buf);
		(*marks)[n].position = pos;
		(*marks)[n].style = GtkCustomRulerMarkMajor;
	}

	return nmarks;
}



