/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: time_axis_view.cc,v 1.15 2003/08/14 04:08:22 pbd Exp $
*/

#include <stdlib.h>
#include <math.h>

#include <algorithm>
#include <string>

#include <list>

#include <pbd/error.h>

#include <gtkmmext/utils.h>
#include <gtkmmext/selector.h>
#include <gtkmmext/gtk_ui.h>
#include <gtkmmext/stop_signal.h>

#include <ardour/session.h>
#include <ardour/utils.h>
#include <ardour/ladspa_plugin.h>
#include <ardour/insert.h>
#include <ardour/location.h>

#include "ardour_ui.h"
#include "editor.h"
#include "time_axis_view.h"
#include "canvas-simplerect.h"
#include "selection.h"

#include "i18n.h"

using namespace Gtk;
using namespace SigC;

TimeAxisView::TimeAxisView(ARDOUR::Session& sess, Editor& ed, Gtk::Widget *canvas) 
	: AxisView(sess), 
	  editor(ed),
	  expansion_arrow (GTK_ARROW_DOWN, GTK_SHADOW_OUT)
{
	canvas_display = gtk_canvas_item_new (gtk_canvas_root(GTK_CANVAS(canvas->gtkobj())),
					      gtk_canvas_group_get_type(),
					      "x", 0.0,
					      "y", 0.0,
					      NULL);
   	control_parent = 0;
	display_menu = 0;
	size_menu = 0;
	hidden = false;
	selected = false;
	height = 0;

	/*
	  Create the standard LHS Controls
	  We create the top-level container and name add the name label here,
	  subclasses can add to the layout as required
	*/

	name_entry.set_name ("EditorTrackNameDisplay");
	name_entry.button_release_event.connect (slot (*this, &TimeAxisView::name_entry_button_release));
	name_entry.button_press_event.connect (slot (*this, &TimeAxisView::name_entry_button_press));
	
	name_entry.focus_in_event.connect (slot (ARDOUR_UI::generic_focus_in_event));
	name_entry.focus_out_event.connect (slot (ARDOUR_UI::generic_focus_out_event));

	Gtkmmext::set_usize_to_display_given_text (name_entry, N_("gTortnam"), 10, 10); // just represents a short name

	name_label.set_name ("TrackLabel");
	name_label.set_alignment (0.0, 0.5);

	expansion_arrow.set_name ("TrackExpansionArrow");
	expansion_arrow.show ();

	expansion_box.set_name ("TimeAxisViewControlsBaseUnselected");
	expansion_box.add (expansion_arrow);
	expansion_box.set_usize (9, 9);
	expansion_box.set_events (GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);
	expansion_box.button_release_event.connect (slot (*this, &TimeAxisView::expansion_button_release_handler));

	// name_hbox.set_border_width (2);
	// name_hbox.set_spacing (5);

	/* typically, either name_label+expansion_box OR name_entry are visible,
	   but not both. its up to derived classes to show/hide them as they
	   wish.
	*/

	name_hbox.pack_start (name_label, true, true);
	
	VBox* vb = manage (new VBox);
	vb->pack_start (expansion_box, false, false);
	vb->show ();

	name_hbox.pack_start (*vb, false, false);
	name_hbox.pack_start (name_entry, true, true);
	name_hbox.show ();

	name_vbox.pack_start (name_hbox, false, false);
	name_vbox.show ();

	controls_hbox.set_border_width (2);
	controls_hbox.pack_start (name_vbox, true, true);
	controls_hbox.show ();

	controls_vbox.pack_start (controls_hbox, false, false);
	controls_vbox.show ();

	controls_ebox.set_name ("TimeAxisViewControlsBaseUnselected");
	controls_ebox.add (controls_vbox);
	controls_ebox.add_events (GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK);
	controls_ebox.set_flags (GTK_CAN_FOCUS);

	controls_frame.add (controls_ebox);
	controls_frame.set_name ("TimeAxisViewControlsBaseUnselected");
	controls_frame.set_shadow_type (GTK_SHADOW_OUT);
}

TimeAxisView::~TimeAxisView()
{
	if (canvas_display) {
		gtk_object_destroy (GTK_OBJECT (canvas_display));	
		canvas_display = 0;
	}
	if (display_menu) {
		delete display_menu;
		display_menu = 0;
	}
	if (size_menu) {
		delete size_menu;
		size_menu = 0;
	}
}

guint32
TimeAxisView::show_at (double y, int nth, VBox *parent)
{
	gdouble ix1, ix2, iy1, iy2;

	if (control_parent) {
		control_parent->reorder_child (controls_frame, nth);
	} else {
		control_parent = parent;
		parent->pack_start (controls_frame, false, false);
		parent->reorder_child (controls_frame, nth);

		controls_frame.show ();
		controls_ebox.show ();
		controls_ebox.queue_resize();
		controls_ebox.button_release_event.connect (slot (*this, &TimeAxisView::controls_ebox_button_release));
	}

	/* the coordinates used here are in the system of the
	   item's parent ...
	*/

	gtk_canvas_item_get_bounds (canvas_display, &ix1, &iy1, &ix2, &iy2);
	gtk_canvas_item_i2w (canvas_display->parent, &ix1, &iy1);
	if (iy1 < 0) {
		iy1 = 0;
	}
	gtk_canvas_item_move (canvas_display, 0.0, y - iy1);
	gtk_canvas_item_show (canvas_display); /* XXX not necessary */

	y_position = y;

	return height;
}

gint
TimeAxisView::controls_ebox_button_release (GdkEventButton* ev)
{
	if (ev->button == 1) {
		if (Keyboard::modifier_state_equals (ev->state, Keyboard::Control)) {
			// editor.set_selected_mixer_strip (0);
		} else {
			editor.set_selected_mixer_strip (*this);
		}
		return TRUE;
	}
	return FALSE;
}



void
TimeAxisView::hide ()
{
	gtk_canvas_item_hide (canvas_display);
	controls_frame.hide ();

	if (control_parent) {
		control_parent->remove (controls_frame);
		control_parent = 0;
	}

	y_position = -1;
}

void
TimeAxisView::step_height (bool bigger)
{
	switch (height) {
	case Largest:
		if (!bigger) set_height (Large);
		break;
	case Large:
		if (bigger) set_height (Largest);
		else set_height (Larger);
		break;
	case Larger:
		if (bigger) set_height (Large);
		else set_height (Normal);
		break;
	case Normal:
		if (bigger) set_height (Larger);
		else set_height (Smaller);
		break;
	case Smaller:
		if (bigger) set_height (Normal);
		else set_height (Small);
		break;
	case Small:
		if (bigger) set_height (Smaller);
		break;
	}
}


void
TimeAxisView::set_height (TrackHeight h)
{
	height = (guint32) h;
	controls_frame.set_usize (-1, height);
}


gint
TimeAxisView::name_entry_button_press (GdkEventButton *ev)
{
	if (ev->button == 3) {
		return do_not_propagate (ev);
	}
	return FALSE;
}

gint
TimeAxisView::name_entry_button_release (GdkEventButton *ev)
{
	if (ev->button == 3) {
		popup_display_menu (ev->time);
		return stop_signal (name_entry, "button_release_event");
	}
	return FALSE;
}

void
TimeAxisView::popup_display_menu (guint32 when)
{
	if (display_menu == 0) {
		build_display_menu ();
	}
	display_menu->popup (1, when);	
}

gint
TimeAxisView::size_click (GdkEventButton *ev)
{
	popup_size_menu (ev->time);
	return TRUE;
}

void
TimeAxisView::popup_size_menu (guint32 when)
{
	if (size_menu == 0) {
		build_size_menu ();
	}
	size_menu->popup (1, when);
}

void
TimeAxisView::set_selected (bool yn)
{
	if (yn) {
		expansion_box.set_name (controls_base_selected_name);
		controls_ebox.set_name (controls_base_selected_name);
		controls_frame.set_name (controls_base_selected_name);
	} else {
		expansion_box.set_name (controls_base_unselected_name);
		controls_ebox.set_name (controls_base_unselected_name);
		controls_frame.set_name (controls_base_unselected_name);
	}
}

void
TimeAxisView::build_size_menu ()
{
	using namespace Menu_Helpers;

	size_menu = new Menu;
	
	MenuList& items = size_menu->items();
	
	items.push_back (MenuElem (_("Largest"), bind (slot (this, &TimeAxisView::set_height), Largest)));
	items.push_back (MenuElem (_("Large"), bind (slot (this, &TimeAxisView::set_height), Large)));
	items.push_back (MenuElem (_("Larger"), bind (slot (this, &TimeAxisView::set_height), Larger)));
	items.push_back (MenuElem (_("Normal"), bind (slot (this, &TimeAxisView::set_height), Normal)));
	items.push_back (MenuElem (_("Smaller"), bind (slot (this, &TimeAxisView::set_height), Smaller)));
	items.push_back (MenuElem (_("Small"), bind (slot (this, &TimeAxisView::set_height), Small)));
}

void
TimeAxisView::build_display_menu ()
{
	using namespace Menu_Helpers;

	display_menu = new Menu;
	display_menu->map_event.connect (slot (*this, &TimeAxisView::handle_display_menu_map_event));

	// Just let implementing classes define what goes into the manu
}

gint
TimeAxisView::expansion_button_release_handler (GdkEventButton* ev)
{
	popup_display_menu (ev->time);
	return TRUE;
}
